package traceroute

import (
	"errors"
	"fmt"
	"golang.org/x/net/icmp"
	"golang.org/x/net/ipv4"
	"golang.org/x/net/ipv6"
	"math/rand"
	"net"
	"time"
)

// from internal/iana/const.go
// Protocol Numbers, Updated: 2017-10-13
const (
	ProtocolIPv6ICMP = 58 // ICMP for IPv6
	ProtocolICMP     = 1  // Internet Control Message
)

// Result holds the data
type Result struct {
	Hop     int
	Station string
	Latency time.Duration
	Err     error
}

// Results is a collection of hops
type Results struct {
	Hops []Result
}

func (s *Session) doV4Hop(i int) Result {
	s.icmpEcho.Body.(*icmp.Echo).Seq = i

	r := Result{
		Hop:     i,
		Station: "*",
	}

	writeBuffer, err := s.icmpEcho.Marshal(nil)

	if err != nil {
		r.Err = err
		return r
	}

	if err := s.ipV4Sock.SetTTL(i); err != nil {
		r.Err = fmt.Errorf("socket: %w", err)
		return r
	}

	timeNow := time.Now()

	dst := s.Destination

	a := net.IPAddr{
		IP:   dst.ip,
		Zone: "",
	}

	if _, err := s.ipV4Sock.WriteTo(writeBuffer, nil, &a); err != nil {
		r.Err = err
		return r
	}

	if err := s.ipV4Sock.SetReadDeadline(time.Now().Add(s.Timeout)); err != nil {
		r.Err = err
		return r
	}

	for ; ; {
		readBytes, _, hopNode, err := s.ipV4Sock.ReadFrom(s.readBuffer)

		if hopNode != nil {
			r.Station = hopNode.String()
		}

		if err != nil {
			r.Err = err
			return r
		}

		icmpAnswer, err := icmp.ParseMessage(ProtocolICMP, s.readBuffer[:readBytes])

		if err != nil {
			r.Err = err
			return r
		}

		latency := time.Since(timeNow)
		r.Latency = latency

		if icmpAnswer.Type == ipv4.ICMPTypeTimeExceeded {
			s.nextHop++
			return r
		}

		if icmpAnswer.Type == ipv4.ICMPTypeEchoReply {
			s.isFinale = true
			return r
		}

	}

}

func (s *Session) doV6Hop(i int) Result {
	s.icmpEcho.Body.(*icmp.Echo).Seq = i

	r := Result{
		Hop:     i,
		Station: "*",
	}

	writeBuffer, err := s.icmpEcho.Marshal(nil)

	if err != nil {
		r.Err = err
		return r
	}

	if err := s.ipV6Sock.SetHopLimit(i); err != nil {
		r.Err = fmt.Errorf("socket: %w", err)
		return r
	}

	timeNow := time.Now()

	dst := s.Destination

	a := net.IPAddr{
		IP:   dst.ip,
		Zone: "",
	}

	if _, err := s.ipV6Sock.WriteTo(writeBuffer, nil, &a); err != nil {
		r.Err = err
		return r
	}

	if err := s.ipV6Sock.SetReadDeadline(time.Now().Add(s.Timeout)); err != nil {
		r.Err = err
		return r
	}

	for ; ; {

		readBytes, _, hopNode, err := s.ipV6Sock.ReadFrom(s.readBuffer)

		if hopNode != nil {
			r.Station = hopNode.String()
		}

		if err != nil {
			r.Err = err
			return r
		}

		icmpAnswer, err := icmp.ParseMessage(ProtocolIPv6ICMP, s.readBuffer[:readBytes])

		if err != nil {
			r.Err = err
			return r
		}

		latency := time.Since(timeNow)
		r.Latency = latency

		if icmpAnswer.Type == ipv6.ICMPTypeTimeExceeded {
			s.nextHop++
			return r
		}

		if icmpAnswer.Type == ipv6.ICMPTypeDestinationUnreachable {
			s.nextHop++
			return r
		}

		if icmpAnswer.Type == ipv6.ICMPTypeEchoReply {
			s.isFinale = true
			return r
		}
	}

}

// TraceRouteV4 run V4 Traceroute
func (s *Session) TraceRouteV4() (*Results, error) {

	sock, err := net.ListenPacket("ip4:icmp", s.Source.ip.String())

	if err != nil {
		return nil, err
	}

	defer sock.Close()
	s.ipV4Sock = ipv4.NewPacketConn(sock)
	defer s.ipV4Sock.Close()

	if err := s.ipV4Sock.SetControlMessage(ipv4.FlagTTL|ipv4.FlagDst|ipv4.FlagInterface|ipv4.FlagSrc, true); err != nil {
		return nil, err
	}

	s.icmpEcho = icmp.Message{
		Type: ipv4.ICMPTypeEcho, Code: 0, Body: &icmp.Echo{ID: rand.Int(), Data: []byte("")},
	}

	return s.doTrace(func(i int) Result {
		return s.doV4Hop(i)
	})
}

func (s *Session) doTrace(cb func(i int) Result) (*Results, error) {
	s.readBuffer = make([]byte, 1500)

	results := Results{}

	for i := 1; i <= s.MaxHops; i++ {
		r := cb(i)
		results.Hops = append(results.Hops, r)
		if s.CallBack != nil {
			s.CallBack(r)
		}

		if s.isFinale {
			break
		}
	}

	return &results, nil
}

// TraceRouteV6 run V6 trace
func (s *Session) TraceRouteV6() (*Results, error) {

	if s.Source.ip==nil {
		return nil, errors.New("missing source ip")
	}
	
	sock, err := net.ListenPacket("ip6:ipv6-icmp", s.Source.ip.String())

	if err != nil {
		return nil, err
	}

	defer sock.Close()
	s.ipV6Sock = ipv6.NewPacketConn(sock)
	defer s.ipV6Sock.Close()

	if err := s.ipV6Sock.SetControlMessage(ipv6.FlagHopLimit|ipv6.FlagDst|ipv6.FlagInterface|ipv6.FlagSrc, true); err != nil {
		return nil, err
	}

	s.icmpEcho = icmp.Message{
		Type: ipv6.ICMPTypeEchoRequest, Code: 0, Body: &icmp.Echo{ID: rand.Int(), Data: []byte("")},
	}

	return s.doTrace(func(i int) Result {
		return s.doV6Hop(i)
	})
}

// TraceRoute measures the steps to the target host 
func (s *Session) TraceRoute() (*Results, error) {

	if s.Destination.isV4() {
		results, err := s.TraceRouteV4()
		if err != nil {
			return nil, err
		}
		return results, nil
	}

	if s.Destination.isV6() {
		results, err := s.TraceRouteV6()
		if err != nil {
			return nil, err
		}
		return results, nil
	}

	return nil, errors.New("could not traceroute " + s.Destination.String())

}
